import tkinter as tk
from threading import Thread, Lock, Condition
import queue
import time


class SharedFile:
    """ Singleton stellt sicher, dass nur ein Datei-Objekt existiert. """
    _instance = None
    """ Lock für Synchronisierung """
    _lock = Lock()
    _condition = Condition(_lock)

    def __init__(self):
        """ aktueller Dateiinhalt """
        self._inhalt = "Starttext"
        """ aktuelle Anzahl der Leser """
        self._anzahlLeser = 0
        """ ist gerade ein Schreiber aktiv """
        self._schreiberAktiv = False


    """Singleton stellt sicher, dass nur ein Datei-Objekt existiert."""
    @classmethod
    def SharedFileGeben(cls):
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance

    """
    Ein Leser beginnt zu lesen.
    :return: Dateiinhalt
    """
    def LesenBeginnen(self):
        # Hier ergänzen
        pass

    """
    Ein Leser beendet das Lesen.
    """
    def LesenBeenden(self):
        # Hier ergänzen
        pass
    
    """
    Ein Schreiber beginnt zu schreiben.
    :return: Dateiinhalt
    """
    def SchreibenBeginnen(self):
        # Hier ergänzen
        pass
    """
    Ein Schreiber speichert vorgenommene Änderungen.
    :param inhaltNeu neuer Inhalt
    """
    def ÄnderungenSpeichern(self, inhaltNeu):
        self._inhalt = inhaltNeu

    """
    Ein Schreiber beendet das Schreiben.
    """
    def SchreibenBeenden(self):
        # Hier ergänzen
        pass



class Schreiber(Thread):
    """
    Schreiber in eine gemeinsame Datei.
    """
    def __init__(self, fenster, update_queue):
        super().__init__()
        self.fenster = fenster
        self.update_queue = update_queue
        self.editierbarerText = tk.Text(self.fenster)
        self.editierbarerText.grid(row=0, column=0, sticky="nsew", padx=5, pady=5)
        self.übernehmen = tk.Button(self.fenster, text="Änderungen speichern", command=self.ÄnderungenSpeichern)
        self.übernehmen.grid(row=1, column=0, sticky="ew", padx=5)
        self.fenster.grid_rowconfigure(0, weight=1)
        self.fenster.grid_columnconfigure(0, weight=1)
        self.fenster.protocol("WM_DELETE_WINDOW", self.on_close)

        self.update_text(SharedFile.SharedFileGeben().SchreibenBeginnen())
        SharedFile.SharedFileGeben().SchreibenBeenden()

    def run(self):
        while True:
            aktuellerInhalt = SharedFile.SharedFileGeben().SchreibenBeginnen()
            aktuellerText = self.editierbarerText.get("1.0", tk.END).strip()
            if aktuellerInhalt != aktuellerText:
                self.update_queue.put(lambda: self.update_text(aktuellerText))
            time.sleep(100)

    def update_text(self, text):
        self.editierbarerText.delete(1.0, tk.END)
        self.editierbarerText.insert(tk.END, text)

    def ÄnderungenSpeichern(self):
        inhalt = self.editierbarerText.get(1.0, tk.END)
        SharedFile.SharedFileGeben().ÄnderungenSpeichern(inhalt.strip())
        SharedFile.SharedFileGeben().SchreibenBeenden()

    def on_close(self):
        self.ÄnderungenSpeichern()
        self.fenster.destroy()



class Leser(Thread):
    """
    Leser in eine gemeinsame Datei.
    """
    def __init__(self, fenster, update_queue):
        super().__init__()
        self.fenster = fenster
        self.update_queue = update_queue
        self.ausgabeText = tk.Text(self.fenster, bg="lightgray", fg="black", state=tk.DISABLED)
        self.ausgabeText.pack(fill="both", expand=True)
        self.fenster.protocol("WM_DELETE_WINDOW", self.on_close)

    def run(self):
        while True:
            aktuellerInhalt = SharedFile.SharedFileGeben().LesenBeginnen()
            self.update_queue.put(lambda: self.update_text(aktuellerInhalt))
            time.sleep(1)

    def update_text(self, text):
        self.ausgabeText.config(state=tk.NORMAL)
        self.ausgabeText.delete(1.0, tk.END)
        self.ausgabeText.insert(tk.END, text)
        self.ausgabeText.config(state=tk.DISABLED)

    def on_close(self):
        SharedFile.SharedFileGeben().LesenBeenden()
        self.fenster.destroy()

# notwendig, um die GUI-Elemente zu aktualisieren
def process_queue(root, gui_queue):
    try:
        while True:
            update = gui_queue.get_nowait()
            update()
    except queue.Empty:
        pass
    root.after(100, process_queue, root, gui_queue)


if __name__ == "__main__":
    gui_queue = queue.Queue()
    root = tk.Tk()
    root.title("Leser und Schreiber Vorlage")
    root.geometry("100x0") # dieses Fenster ist absichtlich nicht sichtbar

    top_frame = tk.Toplevel(root)
    top_frame.title("Schreiber")
    top_frame.geometry("400x300")
    schreiber = Schreiber(top_frame, gui_queue)
    schreiber.start()

    leser_frame = tk.Toplevel(root)
    leser_frame.title("Leser")
    leser_frame.geometry("400x300")
    leser = Leser(leser_frame, gui_queue)
    leser.start()

    leser2_frame = tk.Toplevel(root)
    leser2_frame.title("Leser 2")
    leser2_frame.geometry("400x300")
    leser2 = Leser(leser2_frame, gui_queue)
    leser2.start()

    root.after(100, process_queue, root, gui_queue)
    root.mainloop()
